#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <stdint.h>
#include <fcntl.h>
#include <stdbool.h>
#include <err.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static char hex_digit(unsigned char half_byte) {
  if (half_byte <= 9)
    return '0' + half_byte;
  else
    return 'a' + (half_byte - 10);
}

static void hexdump(void *_data, size_t byte_count, size_t align_bytes) {
  printf("hexdump(%p, 0x%lx, %zu)\n", _data, (unsigned long)byte_count, align_bytes);
  bool last_was_all_zeroes = false;
  size_t byte_count_with_align = byte_count + align_bytes;
  unsigned char *data_aligned_down = _data - align_bytes;
  for (long byte_offset = 0; byte_offset < byte_count_with_align;) {
    unsigned char *bytes = data_aligned_down + byte_offset;
    unsigned long line_bytes = (byte_count_with_align - byte_offset > 16) ?
            16 : (byte_count_with_align - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    bool all_zeroes = true;
    for (int i=0; i<16; i++) {
      if (byte_offset+i < align_bytes || i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
        if (bytes[i] != 0)
          all_zeroes = false;
      }
    }

    if (all_zeroes) {
      if (!last_was_all_zeroes)
        puts("[ zeroes ]");
      last_was_all_zeroes = true;
    } else {
      last_was_all_zeroes = false;

      linep += sprintf(linep, " |");
      for (int i=0; i<16; i++) {
        if (byte_offset+i < align_bytes || i >= line_bytes) {
          *(linep++) = ' ';
          continue;
        }
        if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
          *(linep++) = bytes[i];
        } else {
          *(linep++) = '.';
        }
      }
      linep += sprintf(linep, "|");

      for (long hex_off = 0; hex_off < 16; hex_off += 8) {
        linep += sprintf(linep, "  0x");
        for (int i=hex_off+7; i>=hex_off; i--) {
          if (byte_offset+i < align_bytes || i >= line_bytes) {
            *(linep++) = '?';
            *(linep++) = '?';
            continue;
          }
          *(linep++) = hex_digit(bytes[i]>>4);
          *(linep++) = hex_digit(bytes[i]&0xf);
        }
      }

      puts(line);
    }
    byte_offset += 16;
  }
}

typedef	enum tag_img_bool {
	IMG_FALSE		= 0,
	IMG_TRUE		= 1,
	IMG_FORCE_ALIGN = 0x7FFFFFFF
} IMG_BOOL;
typedef char IMG_CHAR;
typedef uint64_t IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint8_t IMG_UINT8;
typedef uint32_t IMG_UINT32;
typedef uint64_t IMG_UINT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;
typedef int32_t PVRSRV_FENCE;
typedef int32_t PVRSRV_TIMELINE;
typedef IMG_UINT32      IMG_PID;
#define RGXFW_ALIGN			__attribute__ ((aligned (8)))

struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD  0xc0206440

static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
	uint32_t init_module;
};
static void srvkm_init(void) {
  struct drm_pvr_srvkm_init_data data = { .init_module = PVR_SRVKM_SERVICES_INIT };
  SYSCHK(ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

#define PVRSRV_BRIDGE_SYNC 2

#define PVRSRV_BRIDGE_SYNC_ALLOCSYNCPRIMITIVEBLOCK 0
typedef struct PVRSRV_BRIDGE_IN_ALLOCSYNCPRIMITIVEBLOCK_TAG
{
	//IMG_UINT32 ui32EmptyStructPlaceholder;
} __packed PVRSRV_BRIDGE_IN_ALLOCSYNCPRIMITIVEBLOCK;
typedef struct PVRSRV_BRIDGE_OUT_ALLOCSYNCPRIMITIVEBLOCK_TAG
{
	IMG_HANDLE hSyncHandle;
	IMG_HANDLE hhSyncPMR;
	PVRSRV_ERROR eError;
	IMG_UINT32 ui32SyncPrimBlockSize;
	IMG_UINT32 ui32SyncPrimVAddr;
} __packed PVRSRV_BRIDGE_OUT_ALLOCSYNCPRIMITIVEBLOCK;

#define PVRSRV_BRIDGE_SYNCTRACKING 23

#define PVRSRV_BRIDGE_SYNCTRACKING_SYNCRECORDADD 1
typedef struct PVRSRV_BRIDGE_IN_SYNCRECORDADD_TAG
{
	IMG_HANDLE hhServerSyncPrimBlock;
	const IMG_CHAR *puiClassName;
	IMG_BOOL bbServerSync;
	IMG_UINT32 ui32ClassNameSize;
	IMG_UINT32 ui32ui32FwBlockAddr;
	IMG_UINT32 ui32ui32SyncOffset;
} __packed PVRSRV_BRIDGE_IN_SYNCRECORDADD;
typedef struct PVRSRV_BRIDGE_OUT_SYNCRECORDADD_TAG
{
	IMG_HANDLE hhRecord;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_SYNCRECORDADD;

#define PVRSRV_BRIDGE_SRVCORE 1
#define PVRSRV_BRIDGE_SRVCORE_DUMPDEBUGINFO 7
typedef struct PVRSRV_BRIDGE_IN_DUMPDEBUGINFO_TAG
{
#define DEBUG_REQUEST_VERBOSITY_LOW     0
#define DEBUG_REQUEST_VERBOSITY_MEDIUM  1
#define DEBUG_REQUEST_VERBOSITY_HIGH    2
	IMG_UINT32 ui32VerbLevel;
} __packed PVRSRV_BRIDGE_IN_DUMPDEBUGINFO;
typedef struct PVRSRV_BRIDGE_OUT_DUMPDEBUGINFO_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DUMPDEBUGINFO;


#define PVRSRV_BRIDGE_DI 25

#define PVRSRV_BRIDGE_DI_DICREATECONTEXT 0
typedef struct PVRSRV_BRIDGE_IN_DICREATECONTEXT_TAG
{
	IMG_CHAR *puiStreamName;
} __packed PVRSRV_BRIDGE_IN_DICREATECONTEXT;
typedef struct PVRSRV_BRIDGE_OUT_DICREATECONTEXT_TAG
{
	IMG_HANDLE hContext;
	IMG_CHAR *puiStreamName;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DICREATECONTEXT;

#define PVRSRV_BRIDGE_DI_DILISTALLENTRIES 4
typedef struct PVRSRV_BRIDGE_IN_DILISTALLENTRIES_TAG
{
	IMG_HANDLE hContext;
} __packed PVRSRV_BRIDGE_IN_DILISTALLENTRIES;
typedef struct PVRSRV_BRIDGE_OUT_DILISTALLENTRIES_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DILISTALLENTRIES;

#define PVRSRV_BRIDGE_DI_DIREADENTRY 2
typedef struct PVRSRV_BRIDGE_IN_DIREADENTRY_TAG
{
	IMG_UINT64 ui64Offset;
	IMG_UINT64 ui64Size;
	IMG_HANDLE hContext;
	const IMG_CHAR *puiEntryPath;
} __packed PVRSRV_BRIDGE_IN_DIREADENTRY;
typedef struct PVRSRV_BRIDGE_OUT_DIREADENTRY_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DIREADENTRY;


#define PVRSRV_BRIDGE_PVRTL 15

#define PVRSRV_BRIDGE_PVRTL_TLOPENSTREAM 0
typedef struct PVRSRV_BRIDGE_IN_TLOPENSTREAM_TAG
{
	uint64_t/* const IMG_CHAR * */ puiName;
	IMG_UINT32 ui32Mode;
} __packed PVRSRV_BRIDGE_IN_TLOPENSTREAM;
typedef struct PVRSRV_BRIDGE_OUT_TLOPENSTREAM_TAG
{
	IMG_HANDLE hSD;
	IMG_HANDLE hTLPMR;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_TLOPENSTREAM;

#define PVRSRV_BRIDGE_PVRTL_TLACQUIREDATA 2
typedef struct PVRSRV_BRIDGE_IN_TLACQUIREDATA_TAG
{
	IMG_HANDLE hSD;
} __packed PVRSRV_BRIDGE_IN_TLACQUIREDATA;
typedef struct PVRSRV_BRIDGE_OUT_TLACQUIREDATA_TAG
{
#if NEWEST_POWERVR
	PVRSRV_ERROR eError;
	IMG_UINT32 ui32ReadLen;
	IMG_UINT32 ui32ReadOffset;
#else
	IMG_UINT32 ui32ReadOffset;
	IMG_UINT32 ui32ReadLen;
	PVRSRV_ERROR eError;
#endif
} __packed PVRSRV_BRIDGE_OUT_TLACQUIREDATA;

#define PVRSRV_BRIDGE_PVRTL_TLRELEASEDATA 3
typedef struct PVRSRV_BRIDGE_IN_TLRELEASEDATA_TAG
{
#if NEWEST_POWERVR
	IMG_HANDLE hSD;
	IMG_UINT32 ui32ReadLen;
	IMG_UINT32 ui32ReadOffset;
#else
	IMG_HANDLE hSD;
	IMG_UINT32 ui32ReadOffset;
	IMG_UINT32 ui32ReadLen;
#endif
} __packed PVRSRV_BRIDGE_IN_TLRELEASEDATA;
typedef struct PVRSRV_BRIDGE_OUT_TLRELEASEDATA_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_TLRELEASEDATA;


#define PVRSRV_BRIDGE_MM 6

#define PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE 3
typedef struct PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE_TAG
{
	IMG_HANDLE hBuffer;
} __packed PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE;
typedef struct PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE_TAG
{
	IMG_HANDLE hExtMem;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE;

#define PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR 6
typedef struct PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR_TAG
{
	IMG_HANDLE hExtHandle;
} __packed PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR;
typedef struct PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR_TAG
{
#if NEWEST_POWERVR
	IMG_DEVMEM_ALIGN_T uiAlign;
	IMG_DEVMEM_SIZE_T uiSize;
	IMG_HANDLE hPMR;
	PVRSRV_ERROR eError;
#else
	IMG_HANDLE hPMR;
	IMG_DEVMEM_SIZE_T uiSize;
	IMG_DEVMEM_ALIGN_T uiAlign;
	PVRSRV_ERROR eError;
#endif
} __packed PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR;


#define PVRSRV_BRIDGE_RGXHWPERF 134

#define PVRSRV_BRIDGE_RGXHWPERF_RGXCTRLHWPERF 0
typedef struct PVRSRV_BRIDGE_IN_RGXCTRLHWPERF_TAG
{
#if NEWEST_POWERVR
	IMG_UINT64 ui64Mask;
	IMG_BOOL bToggle;
	IMG_UINT32 ui32StreamId;
#else
	IMG_UINT32 ui32StreamId;
	IMG_BOOL bToggle;
	IMG_UINT64 ui64Mask;
#endif
} __packed PVRSRV_BRIDGE_IN_RGXCTRLHWPERF;
typedef struct PVRSRV_BRIDGE_OUT_RGXCTRLHWPERF_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCTRLHWPERF;


#define PVRSRV_BRIDGE_SYNC 2

#define PVRSRV_BRIDGE_SYNC_SYNCALLOCEVENT 7
typedef struct PVRSRV_BRIDGE_IN_SYNCALLOCEVENT_TAG
{
#if NEWEST_POWERVR
	const IMG_CHAR *puiClassName;
	IMG_BOOL bServerSync;
	IMG_UINT32 ui32ClassNameSize;
	IMG_UINT32 ui32FWAddr;
#else
	IMG_BOOL bServerSync;
	IMG_UINT32 ui32FWAddr;
	IMG_UINT32 ui32ClassNameSize;
	/*const IMG_CHAR **/uint64_t puiClassName;
#endif
} __packed PVRSRV_BRIDGE_IN_SYNCALLOCEVENT;
typedef struct PVRSRV_BRIDGE_OUT_SYNCALLOCEVENT_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_SYNCALLOCEVENT;


static IMG_HANDLE export_to_import(IMG_HANDLE mem_export) {
  PVRSRV_BRIDGE_IN_PMRMAKELOCALIMPORTHANDLE in = { .hBuffer = mem_export };
  PVRSRV_BRIDGE_OUT_PMRMAKELOCALIMPORTHANDLE out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRMAKELOCALIMPORTHANDLE);
  return out.hExtMem;
}

static IMG_HANDLE import_to_pmr(IMG_HANDLE mem_import, size_t *size) {
  PVRSRV_BRIDGE_IN_PMRLOCALIMPORTPMR in = { .hExtHandle = mem_import };
  PVRSRV_BRIDGE_OUT_PMRLOCALIMPORTPMR out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRLOCALIMPORTPMR);
  if (size)
    *size = out.uiSize;
  return out.hPMR;
}

#define PVRSRVTL_PACKETHDR_RESERVED 0x31504C54U
typedef struct {
	IMG_UINT32 uiTypeSize;
	IMG_UINT32 uiReserved;
} PVRSRVTL_PACKETHDR;
#define PVRSRV_SYNC_NAME_LENGTH 32U
typedef union
{
	/*! Data for TYPE_TIMELINE (*Deprecated*). This sub-event is no longer
     *  generated in the HOST stream. Timeline data is now provided in the
     *  CLIENT stream instead.
     */
	struct
	{
		IMG_UINT32 uiPid;             /*!< Identifier of owning process */
    uint32_t leak_pad;
		IMG_UINT64 ui64Timeline_UID1; /*!< Unique identifier for timeline resource */
		IMG_CHAR acName[PVRSRV_SYNC_NAME_LENGTH];
		                              /*!< Label or name given to the sync resource */
		IMG_UINT32 ui32Padding;       /*!< Reserved. Align structure size to 8 bytes */
	} sTimelineAlloc;

	/*! Data for TYPE_FENCE_PVR */
	struct
	{
		IMG_PID uiPID;                /*!< Identifier of owning process */
		PVRSRV_FENCE hFence;          /*!< Unique identifier for the fence resource */
		IMG_UINT32 ui32CheckPt_FWAddr; /*!< Unique identifier of the check point
		                                backing this fence on the GPU */
		IMG_CHAR acName[PVRSRV_SYNC_NAME_LENGTH];
		                              /*!< Label or name given to the sync resource */
	} sFenceAlloc;

	/*! Data for TYPE_SYNC_CP */
	struct
	{
		IMG_UINT32 ui32CheckPt_FWAddr; /*!< Unique identifier for the check point resource */
		PVRSRV_TIMELINE hTimeline;     /*!< Unique identifier for the timeline resource */
		IMG_PID uiPID;                 /*!< Identifier of owning process */
		PVRSRV_FENCE hFence;           /*!< Unique identifier for the fence resource */
		IMG_CHAR acName[PVRSRV_SYNC_NAME_LENGTH];
		                               /*!< Label or name given to the sync resource */
	} sSyncCheckPointAlloc;

	/*! Data for TYPE_FENCE_SW */
	struct
	{
		IMG_PID uiPID;                 /*!< Identifier of owning process */
		PVRSRV_FENCE hSWFence;         /*!< Unique identifier for the SWFence resource */
		PVRSRV_TIMELINE hSWTimeline;   /*!< Unique identifier for the timeline resource */
    uint32_t leak_pad;
		IMG_UINT64 ui64SyncPtIndex;    /*!< Sync-pt index where this SW timeline has reached */
		IMG_CHAR acName[PVRSRV_SYNC_NAME_LENGTH];
		                               /*!< Label or name given to the sync resource */
	} sSWFenceAlloc;

	/*! Data for TYPE_SYNC */
	struct
	{
		IMG_UINT32 ui32FWAddr;         /*!< Identifier of sync resource */
		IMG_CHAR acName[PVRSRV_SYNC_NAME_LENGTH];
		                               /*!< Label or name given to the sync resource */
	} sSyncAlloc;
} RGX_HWPERF_HOST_ALLOC_DETAIL;
typedef enum
{
	RGX_HWPERF_HOST_RESOURCE_TYPE_INVALID,   /*!< Invalid */
	RGX_HWPERF_HOST_RESOURCE_TYPE_SYNC,      /*!< SyncPrim */
	RGX_HWPERF_HOST_RESOURCE_TYPE_TIMELINE_DEPRECATED,
	                                         /*!< Timeline resource packets are
	                                           now emitted in client hwperf buffer */
	RGX_HWPERF_HOST_RESOURCE_TYPE_FENCE_PVR, /*!< Fence for use on GPU (SYNC_CP backed) */
	RGX_HWPERF_HOST_RESOURCE_TYPE_SYNC_CP,   /*!< Sync Checkpoint */
	RGX_HWPERF_HOST_RESOURCE_TYPE_FENCE_SW,  /*!< Fence created on SW timeline */

	RGX_HWPERF_HOST_RESOURCE_TYPE_LAST       /*!< End of enumeration */
} RGX_HWPERF_HOST_RESOURCE_TYPE;
typedef struct
{
	RGX_HWPERF_HOST_RESOURCE_TYPE ui32AllocType;
	RGX_HWPERF_HOST_ALLOC_DETAIL RGXFW_ALIGN uAllocDetail;
} RGX_HWPERF_HOST_ALLOC_DATA;
typedef struct
{
#define HWPERF_PACKET_V2B_SIG		0x48575042
	IMG_UINT32  ui32Sig;        /*!< Always the value HWPERF_PACKET_SIG */
	IMG_UINT32  ui32Size;       /*!< Overall packet size in bytes */
	IMG_UINT32  eTypeId;        /*!< Event type information field */
	IMG_UINT32  ui32Ordinal;    /*!< Sequential number of the packet */
	IMG_UINT64  ui64Timestamp;  /*!< Event timestamp */
} RGX_HWPERF_V2_PACKET_HDR;

void print_stream_contents(char *area, size_t area_len, size_t real_area_len) {
  char *name = NULL;
  for (size_t off = 0; off + sizeof(PVRSRVTL_PACKETHDR) <= area_len; ) {
    name = NULL;
    PVRSRVTL_PACKETHDR *hdr = (void*)(area + off);
    if (hdr->uiReserved != PVRSRVTL_PACKETHDR_RESERVED) {
      printf("<<<<<magic mismatch>>>>>\n");
      return;
    }
    uint8_t type = hdr->uiTypeSize >> 24;
    uint16_t size = hdr->uiTypeSize & 0xffff;
    
    off += sizeof(*hdr);
    char *payload = area + off;
    if (off + size > area_len)
      err(1, "overlong payload");

    size_t want_len = sizeof(RGX_HWPERF_V2_PACKET_HDR)+sizeof(RGX_HWPERF_HOST_ALLOC_DATA);
    if (type == 1 && size <= want_len && off + want_len <= real_area_len) {
      RGX_HWPERF_V2_PACKET_HDR *hdr = (void*)payload;
      if (hdr->ui32Sig != HWPERF_PACKET_V2B_SIG) {
        printf("didn't get HWPERF_PACKET_V2B_SIG ???\n");
      } else if (hdr->eTypeId == (1/*RGX_HWPERF_STREAM_ID1_HOST*/<<16) | (0x03/*RGX_HWPERF_HOST_ALLOC*/)) {
        RGX_HWPERF_HOST_ALLOC_DATA *record = (void*)(hdr+1);
        switch (record->ui32AllocType) {
          case RGX_HWPERF_HOST_RESOURCE_TYPE_SYNC     : name = record->uAllocDetail.sSyncAlloc          .acName; break;
          case RGX_HWPERF_HOST_RESOURCE_TYPE_FENCE_PVR: name = record->uAllocDetail.sFenceAlloc         .acName; break;
          case RGX_HWPERF_HOST_RESOURCE_TYPE_SYNC_CP  : name = record->uAllocDetail.sSyncCheckPointAlloc.acName; break;
          case RGX_HWPERF_HOST_RESOURCE_TYPE_FENCE_SW : name = record->uAllocDetail.sSWFenceAlloc       .acName; break;
          default: break;
        }
        if (record->ui32AllocType == RGX_HWPERF_HOST_RESOURCE_TYPE_FENCE_SW) {
          printf("###############   LEAK PAD: 0x%08x\n", record->uAllocDetail.sSWFenceAlloc.leak_pad);
        }
      } else {
        printf("<<<<<HWPERF packet with type 0x%x>>>>>\n", (hdr->eTypeId&0x7fff));
      }
    } else {
      printf("<<<<<packet type %hhu, size 0x%hx>>>>>\n", type, size);
    }

    off += (size + 7) & ~7;
    //printf("at offset 0x%zx now\n", off);
  }
  if (name) {
    char *name_end = name + PVRSRV_SYNC_NAME_LENGTH;
    char *area_end = area + area_len;
    size_t oob_bytes = (name_end<=area_end)?0:(name_end-area_end);
    char *oob_start = name_end - oob_bytes;
    //printf("trailing record name (%zu bytes OOB):\n", oob_bytes);
    //hexdump(name, PVRSRV_SYNC_NAME_LENGTH);
    printf("OOB data (%zu bytes):\n", oob_bytes);
    hexdump(oob_start, oob_bytes, (unsigned long)oob_start & 0xf);
    printf("\n");
  }
}


int main(void) {
  setbuf(stdout, NULL);
#if NEWEST_POWERVR
  gpu_fd = SYSCHK(open("/dev/dri/by-path/platform-pvr-rogue.1.auto-card", O_RDWR));
  srvkm_init();
#else
  gpu_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));
#endif

  {
    PVRSRV_BRIDGE_IN_RGXCTRLHWPERF in = {
      .ui64Mask = (1<<3/*RGX_HWPERF_HOST_ALLOC*/),
      .ui32StreamId = 1/*RGX_HWPERF_STREAM_ID1_HOST*/
    };
    PVRSRV_BRIDGE_OUT_RGXCTRLHWPERF out;
    do_driver_cmd(PVRSRV_BRIDGE_RGXHWPERF, PVRSRV_BRIDGE_RGXHWPERF_RGXCTRLHWPERF);
  }
  {
    PVRSRV_BRIDGE_IN_SYNCALLOCEVENT in = {
      .puiClassName = (unsigned long)"foo",
      .bServerSync = 0,
      .ui32ClassNameSize = 4,
      .ui32FWAddr = 0x11223344
    };
    PVRSRV_BRIDGE_OUT_SYNCALLOCEVENT out;
    do_driver_cmd(PVRSRV_BRIDGE_SYNC, PVRSRV_BRIDGE_SYNC_SYNCALLOCEVENT);
  }

  IMG_HANDLE tl_stream;
  IMG_HANDLE tl_stream_export;
  {
    PVRSRV_BRIDGE_IN_TLOPENSTREAM in = {
#if NEWEST_POWERVR
      .puiName = (unsigned long)"hwperf_host_129", /*TODO where is this ID from?*/
#else
      .puiName = (unsigned long)"hwperf_host_128",
#endif
      .ui32Mode = 0
    };
    PVRSRV_BRIDGE_OUT_TLOPENSTREAM out;
    do_driver_cmd(PVRSRV_BRIDGE_PVRTL, PVRSRV_BRIDGE_PVRTL_TLOPENSTREAM);
    tl_stream = out.hSD;
    tl_stream_export = out.hTLPMR;
  }
  size_t pmr_size;
  IMG_HANDLE tl_stream_pmr = import_to_pmr(export_to_import(tl_stream_export), &pmr_size);
  printf("tl_stream_pmr = 0x%llx\n", tl_stream_pmr);
  printf("stream mapping size: 0x%zx\n", pmr_size);
  char *stream_mapping = SYSCHK(mmap(NULL, pmr_size, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, tl_stream_pmr<<12));

  time_t last_time = time(NULL);
  while (1) {
    time_t cur_time = time(NULL);
    if (cur_time != last_time) {
      {
        PVRSRV_BRIDGE_IN_SYNCALLOCEVENT in = {
          .puiClassName = (unsigned long)"foo",
          .bServerSync = 0,
          .ui32ClassNameSize = 4,
          .ui32FWAddr = 0x11223344
        };
        PVRSRV_BRIDGE_OUT_SYNCALLOCEVENT out;
        do_driver_cmd(PVRSRV_BRIDGE_SYNC, PVRSRV_BRIDGE_SYNC_SYNCALLOCEVENT);
      }
      last_time = cur_time;
    }

    unsigned int read_off, read_len;
    {
      PVRSRV_BRIDGE_IN_TLACQUIREDATA in = { .hSD = tl_stream };
      PVRSRV_BRIDGE_OUT_TLACQUIREDATA out;
      do_driver_cmd_noerror(PVRSRV_BRIDGE_PVRTL, PVRSRV_BRIDGE_PVRTL_TLACQUIREDATA);
      if (out.eError != 0) {
        if (out.eError == 9/*PVRSRV_ERROR_TIMEOUT*/)
          continue;
        errx(1, "PVRSRV_BRIDGE_PVRTL_TLACQUIREDATA eError %u", out.eError);
      }
      read_off = out.ui32ReadOffset;
      read_len = out.ui32ReadLen;
    }
    //printf("<<<<< read_off=0x%x, read_len=0x%x >>>>>\n", read_off, read_len);
    assert(read_off + read_len <= pmr_size);
    void *read_ptr = stream_mapping + read_off;
    //hexdump(read_ptr, read_len);
    print_stream_contents(read_ptr, read_len, pmr_size-read_off);
    //printf("<<<<< end of read >>>>>\n");
    {
      PVRSRV_BRIDGE_IN_TLRELEASEDATA in = {
        .hSD = tl_stream,
        .ui32ReadLen = read_len,
        .ui32ReadOffset = read_off
      };
      PVRSRV_BRIDGE_OUT_TLRELEASEDATA out;
      do_driver_cmd(PVRSRV_BRIDGE_PVRTL, PVRSRV_BRIDGE_PVRTL_TLRELEASEDATA);
    }
  }
}
